from scipy.signal import convolve, fftconvolve, resample_poly
from scipy.signal import resample
import numpy as np


class ADC(object):

    def __call__(self, signal):
        print("ad will be implemented")
        print("after ad, the sample in fiber will be resampled to signal.sps")
        print("and the signal.sps_in_fiber will be set to signal.sps")

        sps_in_fiber = signal.sps_in_fiber
        sps = signal.sps

        N = 1/(sps / sps_in_fiber)
        N = int(N)

        tempx = resample_poly(signal[0], 1, N)
        tempy = resample_poly(signal[1], 1, N)

        new_sample = np.array([tempx, tempy])
        signal.data_sample_in_fiber = new_sample
        signal.sps_in_fiber = signal.sps
        return signal


class DAC(object):

    def __call__(self, signal):
        sps_in_fiber = np.ceil(signal.sps_in_fiber)
        N = sps_in_fiber / signal.sps
        N = int(N)
        from scipy.signal import resample_poly
        # tempx = resample(signal.data_sample[0, :], N)
        # tempy = resample(signal.data_sample[1, :], N)
        tempx = resample_poly(signal.data_sample, up=N, down=1, axis=1)
        signal.data_sample_in_fiber = tempx
        # signal.sps_in_fiber = sps_in_fiber


class PulseShaping(object):
    '''
        Perform Pluse Shaping. need a dict to construct, the construct should contain:
            key:pulse_shaping
                span
                sps
                alpha

    '''

    def __init__(self, **kwargs):
        self.kind = kwargs['kind']
        self.span = kwargs['span']
        self.sps = kwargs['sps']
        self.alpha = kwargs['alpha']
        assert divmod(self.span * self.sps, 2)[1] == 0
        self.number_of_sample = self.span * self.sps
        self.delay = self.span / 2 * self.sps

        self.filter_tap = self.__design_filter()

    def __design_filter(self):
        if self.kind == 'rrc':
            h = PulseShaping.rcosdesign(self.number_of_sample, self.alpha, 1, self.sps)
            return np.atleast_2d(h)

        if self.kind == 'rc':
            print(
                'error, why do you want to design rc filter,the practical use should be two rrc filters,on in transimit'
                'side,and one in receiver side, rrc filter will be designed')

    @staticmethod
    def rcosdesign(N, alpha, Ts, Fs):
        """
        Generates a root raised cosine (RRC) filter (FIR) impulse response.
        Parameters
        ----------
        N : int
            Length of the filter in samples.
        alpha : float
            Roll off factor (Valid values are [0, 1]).
        Ts : float
            Symbol period in seconds.
        Fs : float
            Sampling Rate in Hz.
        Returns
        ---------
        time_idx : 1-D ndarray of floats
            Array containing the time indices, in seconds, for
            the impulse response.
        h_rrc : 1-D ndarray of floats
            Impulse response of the root raised cosine filter.
        """

        T_delta = 1 / float(Fs)

        sample_num = np.arange(N + 1)
        h_rrc = np.zeros(N + 1, dtype=float)

        for x in sample_num:
            t = (x - N / 2) * T_delta
            if t == 0.0:
                h_rrc[x] = 1.0 - alpha + (4 * alpha / np.pi)
            elif alpha != 0 and t == Ts / (4 * alpha):
                h_rrc[x] = (alpha / np.sqrt(2)) * (((1 + 2 / np.pi) *
                                                    (np.sin(np.pi / (4 * alpha)))) + (
                                                           (1 - 2 / np.pi) * (np.cos(np.pi / (4 * alpha)))))
            elif alpha != 0 and t == -Ts / (4 * alpha):
                h_rrc[x] = (alpha / np.sqrt(2)) * (((1 + 2 / np.pi) *
                                                    (np.sin(np.pi / (4 * alpha)))) + (
                                                           (1 - 2 / np.pi) * (np.cos(np.pi / (4 * alpha)))))
            else:
                h_rrc[x] = (np.sin(np.pi * t * (1 - alpha) / Ts) +
                            4 * alpha * (t / Ts) * np.cos(np.pi * t * (1 + alpha) / Ts)) / \
                           (np.pi * t * (1 - (4 * alpha * t / Ts) * (4 * alpha * t / Ts)) / Ts)

        return h_rrc / np.sqrt(np.sum(h_rrc * h_rrc))

    def rrcfilter(self, signal_interface):
        '''

        :param signal_interface: signal object to be pulse shaping,because a reference of signal object is passed
        so the filter is in place
        :return: None

        '''

        # print("---begin pulseshaping ---,the data sample will be set")
        # upsample by insert zeros

        signal_interface.data_sample = np.zeros(
            (signal_interface.symbol.shape[0], signal_interface.sps * signal_interface.symbol_length))
        signal_interface.data_sample = np.asarray(signal_interface.data_sample, dtype=np.complex)
        for i in range(signal_interface.symbol.shape[0]):
            signal_interface.data_sample[i, :] = signal_interface.upsample(signal_interface.symbol[i, :],
                                                                           signal_interface.sps)[0, :]

        temp = []
        for i in range(signal_interface.data_sample.shape[0]):
            temp.append(convolve(self.filter_tap[0, :], signal_interface.data_sample[i, :]))

        # tempy = convolve(self.filter_tap[0, :], signal_interface.data_sample[1, :])
        # temp_signal = np.array([tempx, tempy])
        temp_signal = np.array(temp)
        # compensate group delay
        temp_signal = np.roll(temp_signal, -int(self.delay), axis=1)

        signal_interface.data_sample = temp_signal[:, :signal_interface.sps * signal_interface.symbol_length]


    def __call__(self, signal):
        self.rrcfilter(signal)


class AWG(object):

    def __init__(self, sps=2, alpha=0.02, span=1024):
        '''

        :param sps: pulse shaping parameters
        :param alpha: reference of the signal object,change signal_interface's property will change all
        :param span

        This function will be used to pulse shaping, the sps is default to 2, it should correspond to signal's sps, then
        the siganl will be resampled to sps_in_fiber automatically

        '''

        # self.signal_interface = signal_interface
        self.roll_off = alpha
        self.span = span
        self.pulse_shaping_filter = PulseShaping(kind='rrc', alpha=alpha, sps=sps, span=span)
        self.dac = DAC()

    def __call__(self, signal_interface):
        self.pulse_shaping_filter.rrcfilter(signal_interface)

        self.dac(signal_interface)
        return signal_interface

    def __str__(self):
        string = f"the pulse shaping filter: roll_off:{self.pulse_shaping_filter.alpha}," \
            f" the sps of pulse shaping filter : {self.signal_interface.sps},  the adc sample per symbol: {self.signal_interface.sps_in_fiber} "

        return string

    def __repr__(self):
        return self.__str__()




if __name__ == '__main__':
    # h = PulseShaping.rcosdesign(32 * 4, 0.2, 1, 4)
    import progressbar
    import time

    bar = progressbar.ProgressBar(max_value=16000)
    for i in range(16000):
        bar.update(i + 1)
        time.sleep(0.01)

    bar.update(16000)
    bar.finish()



'''
    This file contains optical instrument along the signal's propagation
'''
import numpy as np
from numpy.fft import fftfreq
from scipy.constants import h, c
from scipy.fftpack import fft, ifft

from ..Base.SignalInterface import QamSignal, Signal, WdmSignal
from ..Filter.designFilter import LowPassFilter
from typing import List
from scipy.special import erf
import warnings


class Demultiplex(object):

    @staticmethod
    def demux_signal(wdm_signal, signal_index=None, roll_off=0.02, nonlinear_fiber=None):
        '''

        This static function is used to demultiplex a wdm_signal, the signal_index of wdm_signal will be obtained, it
        will be shift to center_frequency, an ideal low pass filter will be used to remove other signal that is not
        interested in.

        :param wdm_signal: the wdm signal to be demulitplexed
        :param signal_index: which signal want to get,return signal objectl,if None.all channel will be returned
        :return:
        '''
        frequences = wdm_signal.relative_frequence
        tarray = np.arange(0, wdm_signal[:].shape[1]) * (1 / wdm_signal.fs_in_fiber)

        # temp = wdm_signal[:]
        temp = np.zeros_like(wdm_signal[:])

        for i in range(temp.shape[0]):
            temp[i, :] = wdm_signal[i, :] * np.exp(-1j * 2 * np.pi * frequences[signal_index] * tarray)

        # center_frequence = wdm_signal.relative_frequence[signal_index]
        pos_freq = wdm_signal.signals[signal_index].symbol_rate / 2 * (1 + roll_off)
        neg_freq = -pos_freq
        # LowPassFilter.inplace_ideal_lowpass(temp,pos_freq,neg_freq,fs_in_fiber=wdm_signal.fs_in_fiber)

        signal = QamSignal(symbol_rate=wdm_signal.signals[signal_index].symbol_rate,
                           sps=wdm_signal.signals[signal_index].sps,
                           sps_in_fiber=wdm_signal.signals[signal_index].sps_in_fiber,
                           is_dp=wdm_signal.signals[signal_index].is_dp, is_from_array=True, is_from_demux=True,
                           mf=wdm_signal.signals[signal_index].mf,
                           symbol_length=wdm_signal.signals[signal_index].symbol_length)

        signal.data_sample_in_fiber = temp
        #
        signal.symbol = wdm_signal.signals[signal_index].symbol
        signal.msg = wdm_signal.signals[signal_index].msg
        # the symbol and msg should not be changed, it is the tx information
        signal.center_frequence = wdm_signal.center_frequence
        LowPassFilter.inplace_ideal_lowpass(signal, pos_freq, neg_freq)
        # signal.symbol = wdm_signal.signals[signal_index].symbol
        return signal


class Multiplex(object):
    '''
        This class implements signal multiplex
    '''

    @staticmethod
    def mux_signal(signals: List[Signal], grid_size):
        '''

        :param signals: signals that will be
        :param grid_size: the grid size of the wdm signal
        :return:
        '''
        fs_in_fiber = []
        for signal in signals:
            fs_in_fiber.append(signal.fs_in_fiber)

        if np.any(np.diff(np.array(fs_in_fiber))):
            assert "the sampling frequence of each signal in fiber should be the same"

        absolute_frequences = []
        for signal in signals:
            absolute_frequences.append(signal.center_frequence)

        absolute_frequences = np.array(absolute_frequences)
        freq = np.min(absolute_frequences) + np.max(absolute_frequences)
        freq = freq / 2

        relative_frequence = np.array(absolute_frequences) - freq
        # 从左到右开始复用
        # 每个wdm信道的采样频率必须保持一致
        # 每个wdm信道的样本点个数也要保持一致,截掉长的采样序列的尾部

        sample_length = []
        for signal in signals:
            sample_length.append(signal.sample_number_in_fiber)

        if np.any(np.diff(sample_length)):
            warnings.warn("the number of sample in fiber are not the same, the longer will be "
                          "trancted")
            min_length = min(sample_length)

            for signal in signals:
                number_to_del = signal.sample_number_in_fiber - min_length
                if number_to_del == 0:
                    continue
                else:
                    signal.data_sample_in_fiber = signal.data_sample_in_fiber[:, 0: - number_to_del]

        sample_number = signals[0].sample_number_in_fiber
        fs = signals[0].fs_in_fiber
        t_array = np.arange(0, sample_number) * (1 / fs)

        wdm_data_sample = 0
        for index, signal in enumerate(signals):
            wdm_data_sample += signal.data_sample_in_fiber * np.exp(
                1j * 2 * np.pi * relative_frequence[index] * t_array)

        wdm_signal = WdmSignal(signals, grid_size, wdm_data_sample)

        return wdm_signal


class Edfa:

    def __init__(self, gain_db, nf, is_ase=True, mode='ConstantGain', expected_power=0):
        '''

        :param gain_db:
        :param nf:
        :param is_ase: 是否添加ase噪声
        :param mode: ConstantGain or ConstantPower
        :param expected_power: 当mode为ConstantPoower  时候，此参数有效
        '''

        self.gain_db = gain_db
        self.nf = nf
        self.is_ase = is_ase
        self.mode = mode
        self.expected_power = expected_power

    def one_ase(self, signal,gain_lin=None):
        '''

        :param signal:
        :return:
        '''
        lamb = signal.center_wave_length
        if gain_lin is None:
            one_ase = (h * c / lamb) * (self.gain_lin * 10 ** (self.nf_lin / 10) - 1) / 2
        else:
            one_ase = (h * c / lamb) * (gain_lin * 10 ** (self.nf_lin / 10) - 1) / 2
        return one_ase

    @property
    def gain_lin(self):
        return 10 ** (self.gain_db / 10)

    @property
    def nf_lin(self):
        return 10 ** (self.nf / 10)

    def traverse(self, signal):
        if self.is_ase:
            noise = self.one_ase(signal) * signal.fs_in_fiber
            noise_sample = np.random.randn(*(signal[:].shape)) + 1j * np.random.randn(*(signal[:].shape))
            each_pol_power = noise 

            noise_sample = np.sqrt(each_pol_power / 2) * noise_sample

        else:
            noise_sample = 0

        if self.mode == 'ConstantGain':
            signal[:] = np.sqrt(self.gain_lin) * signal[:] + noise_sample
            return

        if self.mode == 'ConstantPower':
#             raise NotImplementedError("Not implemented")
            signal_power = np.mean(np.abs(signal.data_sample[0, :]) ** 2) + np.mean(
                np.abs(signal.data_sample[1, :]) ** 2)
            desired_power_linear = (10 ** (self.expected_power / 10)) / 1000
            linear_gain = desired_power_linear / signal_power
            #
            #
            noise = self.one_ase(signal,linear_gain)*signal.fs_in_fiber
            noise_sample = np.random.randn(*(signal[:].shape))+1j*np.random.randn(*(signal[:].shape))
            noise_sample = np.sqrt(noise/2) * noise_sample
            signal[:] = np.sqrt(linear_gain) * signal[:] + noise_sample

        return signal

    def __call__(self, signal):
        self.traverse(signal)
        return signal

    def __str__(self):

        string = f"Model is {self.mode}\n" \
            f"Gain is {self.gain_db} db\n" \
            f"ase is {self.is_ase}\n" \
            f"noise figure is {self.nf}"
        return string

    def __repr__(self):
        return self.__str__()


class WSS(object):

    def __init__(self, frequency_offset, bandwidth, oft):

        '''

        :param frequency_offset: value away from center [GHz]
        :param bandwidth: 3-db Bandwidth [Ghz]
        :param oft:GHZ
        '''
        self.frequency_offset = frequency_offset
        self.bandwidth = bandwidth
        self.otf = oft
        self.H = None
        self.freq = None

    def traverse(self, signal):

        sample = np.zeros_like(signal[:])
        for i in range(sample.shape[0]):
            sample[i, :] = signal[i, :]

        freq = fftfreq(len(sample[0, :]), 1 / signal.fs_in_fiber)
        freq = freq / 1e9
        self.freq = freq
        self.__get_transfer_function(freq)

        for i in range(sample.shape[0]):
            sample[i, :] = ifft(fft(sample[i, :]) * self.H)

        return sample

    def __call__(self, signal):
        sample = self.traverse(signal)
        signal[:] = sample
        return signal

    def __get_transfer_function(self, freq_vector):
        delta = self.otf / 2 / np.sqrt(2 * np.log(2))

        H =  0.5 * delta * np.sqrt(2 * np.pi) * (
                erf((self.bandwidth / 2 - (freq_vector - self.frequency_offset)) / np.sqrt(2) / delta) - erf(
            (-self.bandwidth / 2 - (freq_vector - self.frequency_offset)) / np.sqrt(2) / delta))

        H = H / np.max(H)

        self.H = H

    def plot_transfer_function(self, freq=None):
        import matplotlib.pyplot as plt
        if self.H is None:
            self.__get_transfer_function(freq)
            self.freq = freq
        index = self.H >0.001
        plt.figure(figsize=(20,6))
        plt.subplot(121)
        plt.scatter(self.freq[index], np.abs(self.H[index]),color='b',marker='o')
        plt.xlabel('GHz')
        plt.ylabel('Amplitude')
        plt.title("without log")
        plt.subplot(122)
        plt.scatter(self.freq[index], 10 * np.log10(np.abs(self.H[index])),color='b',marker='o')
        plt.xlabel('GHz')
        plt.ylabel('Amplitude')
        plt.title("with log")
        plt.show()

    def __str__(self):

        string = f'the center_frequency is {0 + self.frequency_offset}[GHZ] \t\n' \
            f'the 3-db bandwidth is {self.bandwidth}[GHz]\t\n' \
            f'the otf is {self.otf} [GHz] \t\n'
        return string

    def __repr__(self):
        return self.__str__()


class MZM(object):
    pass


class IQ(object):
    pass


class LaserSource(object):
    pass




from setuptools import setup, find_packages

setup(
    name = "ocspy",
    version = "0.0.9",
    keywords = ("pip", "optical communication","coherent optical", "ocspy", "ocspy"),
    description = "simulate optical communication",
    long_description = "simulate optical communication",
    license = "MIT Licence",
    packages=find_packages(),

    url = "",
    author = "nigulasikaochuan",
    author_email = "nigulasikaochuan@sjtu.edu.cn",


    include_package_data = True,
    platforms = "any",
    install_requires = []
)



import pickle
import zlib
from scipy.constants import c
from scipy.signal import welch
import numpy as np


def lamb2freq(lam):
    '''

    :param lam: wavelength [m]
    :return: frequence [Hz]
    '''
    return c / lam


def freq2lamb(freq):
    '''

    :param freq: frequence [Hz]
    :return: lambda:[m]
    '''
    return c / freq


def downsample(signal, sps):
    '''
    downsample along row

    can receive siganl object or ndarray
    :param signal: Signal Object or ndarray
    :param sps:
    :return: signal
    '''
    if not isinstance(signal, np.ndarray):
        sample = signal.data_sample_in_fiber
        after_process = np.zeros_like(signal.symbol)

    else:
        sample = np.atleast_2d(signal)
        length = divmod(sample.shape[1], sps)

        if length[1] != 0:
            raise ("after downsample the sample number should be integer")
        after_process = np.zeros((sample.shape[0], length[0]), dtype=sample.dtype)

    for i in range(sample.shape[0]):
        after_process[i, :] = sample[i, ::sps]

    return after_process


def scatterplot_colorful(signal, down_sample=True):
    pass


def scatterplot(signal, down_sample=True,backend='mplt'):
    '''

    :param signal:      signal object or ndarray
    :param down_sample: if down_sample is true , signal must be signal object
    :param backend:     can choose mplt or pyqt
    :return:
    '''

    if down_sample:
        assert hasattr(signal,'sps_in_fiber')
        symbol = downsample(signal,signal.sps_in_fiber)

    else:
        if hasattr(signal,'decision_symbol'):
            symbol = signal.decision_symbol

        else:
            symbol = signal

    pol_number = signal.shape[0]
    if backend == 'mplt':
        import matplotlib.pyplot as plt
        for i in range(pol_number):
            plt.subplot(1,pol_number, i+1)
            ibranch = symbol[i,:].real
            qbranch = symbol[i,:].imag
            plt.scatter(ibranch,qbranch,marker='o',color='b')
        plt.show()



def spectrum_analyzer(signal,fs=None,backend='mplt'):
    '''

    :param signal: signal object or ndarray
    :return: None
    '''


    if isinstance(signal,np.ndarray):
        assert fs is not None
        sample = signal
    else:
        fs = signal.fs_in_fiber

        sample = signal[:]

    pol_number = sample.shape[0]
    if backend =='mplt':
        import matplotlib.pyplot as plt
        plt.figure(figsize=(20,6))
        for i in range(pol_number):
           
            plt.subplot(1,pol_number,i+1)
            [f,pxx] = welch(sample[i,:],fs,nfft=2048,detrend=False,return_onesided=False)
            plt.plot(f/1e9,10*np.log10(np.abs(pxx)))
            plt.xlabel('Frequency [GHZ]')
            plt.ylabel('Power Spectrem Density [db/Hz]')
        plt.show()

def eyedigram(signal):
    pass


def save_signal(fn, signal, lvl=4, **kwargs):
    """
    Save a signal object using zlib compression

    Parameters
    ----------
    fn : basestring
        filename
    signal : SignalObject
        the signal to save
    lvl : int, optional
        the compression to use for zlib
    """
    with open(fn, "wb") as fp:
        sc = zlib.compress(
            pickle.dumps({'signal': signal, 'simulation_information': kwargs}, protocol=pickle.HIGHEST_PROTOCOL),
            level=lvl)
        fp.write(sc)


def load_signal(fn):
    """
    Load a signal object from a zlib compressed pickle file.

    Parameters
    ----------
    fn : basestring
        filename of the file

    Returns
    -------
    sig : SignalObject
        The loaded signal object

    """
    with open(fn, "rb") as fp:
        s = zlib.decompress(fp.read())
        obj = pickle.loads(s)
    return obj
